	integer ITYPE,XMIN,XMAX,YMIN,YMAX,ZMIN,ZMAX,SIZE,DATA,HDR
	parameter(ITYPE=1,XMIN=2,XMAX=3,YMIN=4,YMAX=5,ZMIN=6,ZMAX=7)
	parameter(SIZE=8,DATA=9,HDR=8)
	integer IMAGE,POLY,BOX,QUART,MASK
	integer ABSPIX,RELPIX,RELCEN,ARCSEC,KMS
	parameter(IMAGE=1,POLY=2,BOX=3,QUART=4,MASK=5)
	parameter(ABSPIX=6,RELPIX=7,RELCEN=8,ARCSEC=9,KMS=10)
	integer NX,NY,NZ,OFFSET0
	parameter(NX=2,NY=3,NZ=4,OFFSET0=4)
