c Warning: do not edit this file, it has possibly been generated by configure
c TELESCOPE=carma
c=======================================================================
	CHARACTER*(*) MIRTEL
	PARAMETER(MIRTEL='carma')
	INTEGER   MAXBUF
	PARAMETER(MAXBUF=4194304)
c
	INTEGER   MAXDIM, MAXDIM2
	PARAMETER(MAXDIM=65536,MAXDIM2=8192)
c
	INTEGER   MAXIANT,MAXANT,MAXANT2
	PARAMETER(MAXIANT=256,MAXANT=64,MAXANT2=28)
c		maximum number of baselines
	INTEGER   MAXBASE
c	PARAMETER(MAXBASE=((MAXANT*(MAXANT+1))/2))
	PARAMETER(MAXBASE=500)
	INTEGER   MAXBASE2
	PARAMETER(MAXBASE2=500)
c
	INTEGER   MAXCHAN
	PARAMETER(MAXCHAN=4096)
c
	INTEGER   MAXWIN
	PARAMETER(MAXWIN=16)
c
	INTEGER   MAXWIDE
	PARAMETER(MAXWIDE=18)
c=======================================================================

